#!/usr/bin/env python
# coding: utf-8
"""
metafora - metar
"""
import re
from dataclasses import field, dataclass
from dataclasses_json import dataclass_json, config
from typing import List, Optional
import warnings

warnings.filterwarnings(action="ignore", category=UserWarning)

from metafora.common import (
    Station,
    Timestamp,
    Wind,
    Visibility,
    Clouds,
    Weather,
    Number,
)
from metafora.parser import ParserMixIn
from metafora.utils import convert_distance, convert_pressure, convert_direction


@dataclass_json
@dataclass
class RunwayVisualRange:
    """
    Runway visual range (RVR)
    distances are expressed in meters
    """

    runway: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    distance: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    distance_prefix: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    distance_min: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    distance_max: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    tendency: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the RVR from a token

        :param token: token
        :return: RVR instance or None if not successful
        """
        found = re.search(
            "^R([0-9]{2}[LCR]?)/(([PM])?([0-9]{4})V)?([PM])?([0-9]{4})(FT)?/?([UDN]?)$",
            token,
        )

        if not found:
            return None

        runway = None
        distance = None
        distance_prefix = None
        distance_min = None
        distance_max = None
        tendency = None

        if found[1]:
            runway = found[1]

        unit = "M"
        if found[7]:
            unit = found[7]

        if found[6]:
            if found[4]:
                distance_min = convert_distance(int(found[4]), unit)
                distance_max = convert_distance(int(found[6]), unit)
            else:
                distance = convert_distance(int(found[6]), unit)

            if found[5]:
                distance_prefix = found[5]
        else:
            # if distance is not give, return none
            return None

        if found[8]:
            tendency = found[8]

        return cls(
            runway=runway,
            distance=distance,
            distance_prefix=distance_prefix,
            distance_min=distance_min,
            distance_max=distance_max,
            tendency=tendency,
        )


@dataclass_json
@dataclass
class Temperature:
    """
    Temperature
    Temperature is expressed in degrees celcius
    """

    temperature: Number
    dewpoint: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the temperature from a token

        :param token: token
        :return: visibility conditions instance or None if not successful
        """
        found = re.search("^(M?[0-9]{2})/(M?[0-9]{2}|X{2})?$", token)

        if not found:
            return None

        temperature = int(found[1].replace("M", "-"))

        if found[2] and len(found[2]) != 0 and found[2] != "XX":
            dewpoint = int(found[2].replace("M", "-"))
        else:
            # unknown dew point
            dewpoint = None

        return cls(temperature=temperature, dewpoint=dewpoint)


class Pressure(int):
    """
    Pressure conditions
    Pressure is expressed in hPa
    """

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the pressure from a token

        :param token: token
        :return: pressure conditions instance or None if not successful
        """
        found = re.search("^([QA])(////|[0-9]{4})$", token)

        if not found:
            return None

        if found[2] and found[2] != "////":
            unit = found[1]

            if unit == "A":
                pressure = int(found[2]) / 100
            else:
                pressure = int(found[2])

            pressure = int(convert_pressure(pressure, unit))
        else:
            pressure = None

        return pressure


@dataclass_json
@dataclass
class Metar(ParserMixIn):
    """
    METAR dataclass
    """

    station: Station
    time: Optional[Timestamp] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    wind: Optional[Wind] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    visibility: Optional[Visibility] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    runway_info: Optional[List[RunwayVisualRange]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )
    weather: Optional[List[Weather]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )
    clouds: Optional[List[Clouds]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )
    temperature: Optional[Temperature] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    pressure: Optional[Pressure] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )


def runway_info_distance_min(runway_info: List[RunwayVisualRange]) -> RunwayVisualRange:
    """
    Gets compass of the runway with minimum distance

    :param clouds: list of RVR
    :return: compass and distance
    """    
    distance = None
    compass = None

    if runway_info is not None:
        for r in runway_info:
            if r.distance is not None:
                if distance is None or r.distance < distance:
                    distance = r.distance
                    compass = convert_direction(int(r.runway[:2]))
            elif r.distance_min is not None:
                if distance is None or r.distance_min < distance:
                    distance = r.distance_min
                    compass = convert_direction(int(r.runway[:2]))

    return RunwayVisualRange(runway=compass, distance=distance)


def runway_info_distance_max(runway_info: List[RunwayVisualRange]) -> RunwayVisualRange:
    """
    Gets compass of the runway with maximum distance

    :param clouds: list of RVR
    :return: compass and distance
    """    
    distance = None
    compass = None

    if runway_info is not None:
        for r in runway_info:
            if r.distance is not None:
                if distance is None or r.distance > distance:
                    distance = r.distance
                    compass = convert_direction(int(r.runway[:2]))
            elif r.distance_max is not None:
                if distance is None or r.distance_max > distance:
                    distance = r.distance_max
                    compass = convert_direction(int(r.runway[:2]))
                
    return RunwayVisualRange(runway=compass, distance=distance)



#!/usr/bin/env python
# coding: utf-8
"""
metafora - taf
"""
from dataclasses import field, dataclass
from dataclasses_json import dataclass_json, config
from typing import List, Optional
import re

import warnings

warnings.filterwarnings(action="ignore", category=UserWarning)

from metafora.common import (
    Station,
    Timestamp,
    Wind,
    Clouds,
    Weather,
    Visibility,
    TIME_FORMAT,
    Number
)
from metafora.parser import ParserMixIn, sanitise_string
from metafora.enums import ChangeEu

HOUR_FORMAT = "(0[0-9]|1[0-9]|2[0-9]|3[0-1])(0[0-9]|1[0-9]|2[0-4])"
CHANGE_SEARCH = (
    r"^(PROB[0-9]{2}\sTEMPO)|(TEMPO)|(PROB[0-9]{2})|(INTER)|(BECMG)|(FM"
    + HOUR_FORMAT
    + ")|(AT"
    + HOUR_FORMAT
    + ")|(TL"
    + HOUR_FORMAT
    + ")"
)
CHANGE_SUB = (
    r"(PROB[0-9]{2}\sTEMPO|TEMPO|PROB[0-9]{2}|INTER|BECMG|FM"
    + HOUR_FORMAT
    + "|AT"
    + HOUR_FORMAT
    + "|TL"
    + HOUR_FORMAT
    + ")"
)


@dataclass_json
@dataclass
class Validity:
    """
    Valdity period
    """

    start_time: Optional[Timestamp] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    end_time: Optional[Timestamp] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the validity period information from a token

        :param token: token
        :return: validity instance or None if not successful
        """
        found = re.search(
            "^((AT|FM|TL)"
            + TIME_FORMAT
            + ")|("
            + HOUR_FORMAT
            + "/"
            + HOUR_FORMAT
            + ")$",
            token,
        )

        if not found:
            return None

        start_time = None
        end_time = None

        if found[1]:
            time = Timestamp.from_text(
                str(found[3]) + str(found[4]) + str(found[5]) + "Z"
            )
            if found[2] == "FM":
                start_time = time
                end_time = None
            elif found[2] == "TL":
                start_time = None
                end_time = time
            elif found[2] == "AT":
                start_time = time
                end_time = time
        else:
            start_time = Timestamp.from_text(str(found[7]) + str(found[8]) + "00Z")
            end_time = Timestamp.from_text(str(found[9]) + str(found[10]) + "00Z")

        return cls(start_time=start_time, end_time=end_time)


class Probability(int):
    """
    Probability of change
    """

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the pressure from a token
        :param token: token
        :return: probability of change instance or None if not successful
        """
        found = re.search("^(PROB[0-9]{2})$", token)

        if not found:
            return None

        return int(found[0].replace("PROB", ""))


class Indicator(str):
    """
    Change indicator
    """

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the pressure from a token
        :param token: token
        :return: change indicator instance or None if not successful
        """
        found = re.search("^(" + "|".join(ChangeEu.set()) + ")$", token)

        if not found:
            return None

        return str(found[0])


@dataclass_json
@dataclass
class MaximumTemperature:
    """
    Maximum Temperature
    Maximum Temperature is expressed in degrees celcius
    """

    value: Number
    time: Timestamp

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the temperature from a token
        :param token: token
        :return: visibility conditions instance or None if not successful
        """
        found = re.search("^TX(M?[0-9]{2})/(" + HOUR_FORMAT + ")Z$", token)

        if not found:
            return None

        value = int(found[1].replace("M", "-"))
        time = Timestamp.from_text(str(found[2]) + "00Z")

        return cls(value=value, time=time)


# NOT USED
@dataclass_json
@dataclass
class MinimumTemperature:
    """
    Minimum Temperature
    Minimum Temperature is expressed in degrees celcius
    """

    value: Number
    time: Timestamp

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the temperature from a token
        :param token: token
        :return: visibility conditions instance or None if not successful
        """
        found = re.search("^TN(M?[0-9]{2})/(" + HOUR_FORMAT + ")Z$", token)

        if not found:
            return None

        value = int(found[1].replace("M", "-"))
        time = Timestamp.from_text(str(found[2]) + "00Z")

        return cls(value=value, time=time)
    

@dataclass_json
@dataclass
class Forecast(ParserMixIn):
    """
    Weather forecast dataclass
    """

    probability: Optional[Probability] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    indicator: Optional[Indicator] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    validity: Optional[Validity] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    wind: Optional[Wind] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    visibility: Optional[Visibility] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    weather: Optional[List[Weather]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )
    clouds: Optional[List[Clouds]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )
    maxtemperature: Optional[MaximumTemperature] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    mintemperature: Optional[MinimumTemperature] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    def __post_init__(self):
        if self.validity is None:
            # just for compatibility ... In Python 3.10 we could use kw_only of dataclasses
            raise ValueError("Validity is mandatory in forecasts")

"""         if self.indicator is None:
            if self.validity.start_time is None and self.validity.end_time is not None:
                self.indicator = Indicator("TL")
            elif (
                self.validity.end_time is None and self.validity.start_time is not None
            ):
                self.indicator = Indicator("FM")
            elif (
                self.validity.start_time == self.validity.end_time
                and self.validity.start_time is not None
            ):
                self.indicator = Indicator("AT") """


@dataclass_json
@dataclass
class Taf:
    """
    TAF dataclass
    """

    station: Station
    time: Optional[Timestamp] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    forecasts: Optional[List[Forecast]] = field(
        default_factory=list, metadata=config(exclude=lambda x: len(x) == 0)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the TAF from a token

        :param token: token
        :return: TAF instance or False if not successful
        """
        # sanitise 
        token = sanitise_string(token) 
        
        # check if it follows a TAF format
        found = re.search(
            "^(PROV\\s)?TAF\\s(AMD\\s|COR\\s)?([A-Z]{4})\\s(" + TIME_FORMAT + "Z)?",
            sanitise_string(token),
        )

        if not found:
            raise ValueError("The report '{}' does not look like a TAF".format(token))

        # extract context
        station = Station.from_text(found[3])

        # extract time
        if found[4]:
            time = Timestamp.from_text(found[4])
        else:
            time = None

        # process peridos
        lines = re.sub(CHANGE_SUB, "\n\\g<1>", token.replace(found[0], "")).splitlines()

        # forecasts
        forecasts = [Forecast.from_text(line) for line in lines]

        # create taf instance
        return cls(station=station, time=time, forecasts=forecasts)


def propagate_forecasts(forecasts: List[Forecast]) -> List[Forecast]:
    """
    Propgate forecasts

    :param forecasts:
    :return:
    """
    # convert to dictionaries
    forecasts_dict = [forecast.to_dict() for forecast in forecasts]

    # first is the prevailing
    prevailing = forecasts_dict[0]
    propagated = [prevailing]

    # iterate over forecasts
    for forecast in forecasts_dict[1:]:
        # current is the prevailing conditions
        current = prevailing.copy()

        # get indicator
        indicator = forecast.get("indicator", None)

        # update fields of prevailing weather
        current.update(forecast)

        if indicator in ("BECMG", "FM"):
            current["validity"]["end_time"] = prevailing["validity"]["end_time"].copy()
            prevailing["validity"]["end_time"] = current["validity"][
                "start_time"
            ].copy()
            prevailing = current

        propagated.append(current)

    return [Forecast.from_dict(forecast) for forecast in propagated]



#!/usr/bin/env python
# coding: utf-8
"""
metafora - some helping functions for engineering
"""
from metafora.taf import Forecast, Taf, propagate_forecasts
from metafora.metar import Metar, runway_info_distance_max, runway_info_distance_min
from metafora.common import simplify_clouds
from metafora.enums import WeatherPrecipitation, WeatherObscuration, OtherWeather
from typing import Union, Dict, List, Optional
from pandas import DataFrame, json_normalize, to_datetime, Timestamp, NaT
from datetime import datetime, timedelta

BASIC_COLS = [
              "station", 
              "visibility_cavok", "visibility_distance", 
              "clouds_height", "clouds_amount", "clouds_vertical_visibility", 
              "wind_speed", "wind_gust", "wind_compass",
              ]

BOOLEAN_COLS = ['precipitation', 
                'obscuration',
                'other', 
                'thunderstorms', 
                'freezing',
                'snow', 
                'clouds']

METAR_SPECIFIC_COLS = ["temperature_temperature", 
                       "temperature_dewpoint", 
                       "pressure"]

TAF_SPECIFIC_COLS = ["indicator", "probability", 
                     "maxtemperature_value",
                     "mintemperature_value"]

def simplify_report(report: Union[Metar, Forecast]) -> Dict:
    """
    Simplifies a report (TAF or METAR) and converts the result into a basic dictionary

    :param report: a raw METAR or TAF report
    :return: simplified dictionary representation of the report
    """
    features = report.to_dict()

    # simplify clouds
    features["clouds"] = simplify_clouds(report.clouds).to_dict()

    # simplify runway visual range
    if features.pop("runway_info", None):
        features["runway_info_min"] = runway_info_distance_min(report.runway_info).to_dict()
        features["runway_info_max"] = runway_info_distance_max(report.runway_info).to_dict()

    # flatten weather
    for i, w in enumerate(features.pop("weather", [])):
        features["weather_{}".format(i)] = w

    return features


def time_from_reference(time: Timestamp, reference: datetime) -> datetime:
    """Determines the datetime of a metafora Timestamp from a reference datetime

    Args:
        time (Timestamp): metafora Timestamp
        reference (datetime): reference datetime

    Returns:
        datetime: datetime associated to the timestamp
    """
    # correct minute
    if time.minute == 60:
        time.minute = 0
        time.hour = min(time.hour + 1, 24)

    # correct hour
    if time.hour == 24:
        time.hour = 0
        increased_day = True
    else:
        increased_day = False

    result = reference.replace(hour=time.hour, minute=time.minute)

    # move to next day until they match
    while time.day != result.day:
        result = result + timedelta(days=1)

    # if increased day
    if increased_day:
        result = result + timedelta(days=1)

    return result


def process_tafs(tafs: List[Dict], errors: Optional[str] = "ignore") -> List[Dict]:
    """Processes list of TAFs with metafora
       Each element of the list must be a dictionary with the following keys
       - report: the raw TAF
       - time: the release time

    Args:
        tafs (List[Dict]): list of raw TAFs

    Returns:
        List[Dict]: processed TAFs
    """
    reports = []

    for t in tafs:
        # get release time
        time = t["time"]

        # convert to isoformat
        time_iso = datetime.fromisoformat(time)
        try:
            # parse TAF from raw text
            report = Taf.from_text(t["report"])

            # get station 
            station = report.station

            # unify forecasts
            forecasts = propagate_forecasts(report.forecasts)

            for f in forecasts:
                # get validity period
                start_time = time_from_reference(
                    f.validity.start_time, time_iso)
                end_time = time_from_reference(f.validity.end_time, time_iso)

                # extract ml features
                report = simplify_report(f)

                # set station as well as release time and validity times in isoformat
                report["time"] = time
                report["station"] = station
                report["validity"]["start_time"] = start_time.isoformat()
                report["validity"]["end_time"] = end_time.isoformat()

                # min and max temperatures are special cases ...
                if f.mintemperature is not None and f.mintemperature.time is not None:
                    report["mintemperature"]["time"] = time_from_reference(
                        f.mintemperature.time, time_iso).isoformat()

                if f.maxtemperature is not None and f.maxtemperature.time is not None:
                    report["maxtemperature"]["time"] = time_from_reference(
                        f.maxtemperature.time, time_iso).isoformat()
                
                # append the report
                reports.append(report)
        except Exception:
            if errors == "raise":
                raise ValueError("{} is not a valid TAF".format(t["report"]))

    return reports


def process_metars(metars: List[Dict], errors: Optional[str] = "ignore") -> List[Dict]:
    """Processes list of METARs with metafora
       Each element of the list must be a dictionary with the following keys
       - report: the raw METAR
       - time: the release time

    Args:
        metars (List[Dict]): list of raw METARs

    Returns:
        List[Dict]: processed METARs
    """
    reports = []

    for m in metars:
        try:
            # parse METAR from raw text
            metar = Metar.from_text(m["report"])

            # extract ml features
            report = simplify_report(metar)

            # set release time in isoformat
            report["time"] = m["time"]
            reports.append(report)
        except Exception:
            if errors == "raise":
                raise ValueError("{} is not a valid METAR".format(m["report"]))

    return reports



def machine_learning_features(df: DataFrame) -> DataFrame:
    """Simplifies the weather dataframe
       extracting only the most relevant information for aviation

    Args:
        weather (DataFrame): original weather dataframe

    Returns:
        DataFrame: simplified weather dataframe
    """
    # boolean columns to False by default
    df[BOOLEAN_COLS] = False

    for c in df.columns:
        if c.endswith("phenomena"):
            # precipitation
            status = df[c].str.startswith(
                tuple(WeatherPrecipitation._member_names_)).fillna(False)
            df["precipitation"] = df["precipitation"] | status

            # snow
            status = df[c].str.contains("SN").fillna(False)
            df["snow"] = df["snow"] | status

            # obscuration
            status = df[c].str.startswith(
                tuple(WeatherObscuration._member_names_)).fillna(False)
            df["obscuration"] = df["obscuration"] | status

            # other phenomena
            status = df[c].str.startswith(
                tuple(OtherWeather._member_names_)).fillna(False)
            df["other"] = df["other"] | status
        elif c.endswith("descriptor"):
            # thunderstorms
            status = (df[c] == "TS").fillna(False)
            df["thunderstorms"] = df["thunderstorms"] | status

            # freezing
            status = (df[c] == "FZ").fillna(False)
            df["freezing"] = df["freezing"] | status

    # general clouds (TCU or CB)
    if "clouds_cloud" in df.columns:
        df['clouds'] = df["clouds_cloud"].notnull()

    return df


def reports_to_dataframe(reports: List[Dict], drop_invalid: Optional[bool] = False) -> DataFrame:
    """Converts list of processed reports (METARs or TAFs) to pandas dataframe
    where each row is a report and each column a feature

    Args:
        List (Dict): list of reports processed by metafora
        drop_invalid (Optional[bool], optional): drop invalid reports. Defaults to False.

    Returns:
        DataFrame: reports dataframe
    """
    # json normalize, back to pandas
    df = json_normalize(reports, sep="_")

    # convert time columns to datetime
    TIME_COLS = []
    for c in ["time", "validity_start_time", "validity_end_time", "maxtemperature_time", "mintemperature_time"]:
        if c in df.columns:
            df[c] = to_datetime(df[c], errors="coerce")
            TIME_COLS.append(c)

    # specific columns
    if "validity_start_time" in TIME_COLS and "validity_end_time" in TIME_COLS:
        # TAF case
        SPECIFIC_COLS = TAF_SPECIFIC_COLS
        if drop_invalid:
            df.dropna(subset=["station", "time",
                      "validity_start_time", "validity_end_time"], inplace=True)

        # if max and min temperature times are missing
        for c in ["maxtemperature_time", "mintemperature_time"]:
            if c not in TIME_COLS:
                df[c] = NaT
    else:
        # METAR case
        SPECIFIC_COLS = METAR_SPECIFIC_COLS
        if drop_invalid:
            df.dropna(subset=["station", "time"], inplace=True)

    # cavok must be boolean
    if "visibility_cavok" in df.columns:
        df["visibility_cavok"] = df["visibility_cavok"].fillna(False)
    else:
        df["visibility_cavok"] = False

    # create simplified columns
    df = machine_learning_features(df)

    # enforce basic and specific columns
    for c in BASIC_COLS + SPECIFIC_COLS:
        if c not in df.columns:
            df[c] = None 

    return df[BASIC_COLS + TIME_COLS + BOOLEAN_COLS + SPECIFIC_COLS]



__version__ = "1.1.5"
__author__ = 'Ramon Dalmau-Codina'

from metafora.metar import Metar
from metafora.taf import Taf, Forecast, propagate_forecasts
from metafora.engineering import simplify_report




#!/usr/bin/env python
# coding: utf-8
"""
metafora - parser
"""
from dataclasses import fields, dataclass
from typing import Union, get_origin, get_args
import re


def sanitise_string(raw: str) -> str:
    """
    Remove duplicates spaces, newlines and tabs from string

    :param raw:
    :return:
    """
    sane = " ".join(raw.split())
    return sane.strip().rstrip('=')


@dataclass
class ParserMixIn:
    """
    Text parser
    """

    @classmethod
    def from_text(cls, token: str):
        """
        Parse parts using the fields of a structure

        :param token: raw text
        :return:
        """
        substrings = sanitise_string(token).split(" ")
        class_fields = fields(cls)
        kwargs = dict()

        # current field index
        j = 0

        # for each word
        for s in substrings:
            i = j

            # try to parse this substring
            # until exhausting all class fields
            while i < len(class_fields):
                cf = class_fields[i]

                # get class field type
                cf_type = cf.type

                # get class field name
                cf_name = cf.name

                # get origin of class field type
                if get_origin(cf_type) == Union:
                    cf_origin = get_args(cf_type)[0]
                else:
                    cf_origin = cf_type

                # if it is a list
                if get_origin(cf_origin) == list:
                    is_list = True
                    cf_origin = get_args(cf_origin)[0]
                else:
                    is_list = False

                # the base class can be parsed from text
                if hasattr(cf_origin, "from_text"):
                    result = cf_origin.from_text(s)

                    # parsed
                    if result:
                        if is_list:
                            if cf_name not in kwargs.keys():
                                kwargs[cf_name] = [result]
                            else:
                                kwargs[cf_name].append(result)

                            # keep using this field
                            j = i
                        else:
                            kwargs[cf_name] = result

                            # move to the next field
                            # i = i + 1?
                            j = i + 1

                        # stop
                        break

                # move to next field
                i = i + 1

        return cls(**kwargs)



#!/usr/bin/env python
# coding: utf-8
"""
metafora - common dataclasses
"""
import re
from dataclasses import field, dataclass
from dataclasses_json import dataclass_json, config
from typing import Optional, List, Union

from metafora.utils import (
    convert_distance,
    convert_speed,
    convert_direction,
    Number,
)
from metafora.enums import *

WEATHER_DESCRIPTOR = "|".join(WeatherDescriptor.set())
WEATHER_PHENOMENA = "|".join(WeatherPrecipitation.set().union(WeatherObscuration.set()).union(OtherWeather.set()))

TIME_FORMAT = "(0[0-9]|1[0-9]|2[0-9]|3[0-1])(0[0-9]|1[0-9]|2[0-4])([0-5][0-9])"
DIRECTION_FMT = "0[0-9][0-9]|1[0-9][0-9]|2[0-9][0-9]|3[0-5][0-9]|360"


class Station(str):
    """
    Station class
    """

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the station from a token

        :param token: token
        :return: station instance or None if not successful
        """
        found = re.search("^([A-Z]{4})$", token)

        if not found:
            return None

        return str(found[0])


@dataclass_json
@dataclass
class Timestamp:
    """
    Time information
    """

    day: Number
    hour: Number
    minute: Number

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the time information from a token

        :param token: token
        :return: time instance or None if not successful
        """
        found = re.search("^" + TIME_FORMAT + "Z$", token)

        if not found:
            return None

        day = int(found[1])
        hour = int(found[2])
        minute = int(found[3])

        return cls(day=day, hour=hour, minute=minute)


@dataclass_json
@dataclass
class Wind:
    """
    Wind conditions
    Speed and gust are expressed in meters per second
    """

    speed: Number
    compass: str
    gust: Optional[Number] = field(default=0)

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the wind conditions from a token

        :param token: token
        :return: wind conditions instance or None if not successful
        """
        found = re.search(
            "^("
            + DIRECTION_FMT
            + "|VRB|///)P?([0-9]{2,3}|//)(GP?([0-9]{2,3}))?(KT|MPS|KPH)$",
            token,
        )

        if not found:
            return None

        # get unit
        unit = found[5]

        # get speed
        if found[2] == "//":
            # unknown speed
            speed = None
        else:
            speed = found[2].replace("/", "")
            speed = convert_speed(int(speed), unit)

        # get compass
        if found[1] == "VRB":
            # variable driection
            compass = "VRB"
        elif found[1] == "///":
            # unknown direction
            compass = None
        else:
            direction = int(found[1])
            compass = convert_direction(direction)

        # get gust
        if found[4]:
            gust = convert_speed(int(found[4]), unit)
        else:
            gust = 0

        return cls(speed=speed, compass=compass, gust=gust)


@dataclass_json
@dataclass
class VariableDirection:
    """
    Variable wind direction
    """

    direction_min: str
    direction_max: str

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the wind from a token

        :param token: token
        :return: wind instance or None if not successful
        """
        found = re.search("^(" + DIRECTION_FMT + ")V(" + DIRECTION_FMT + ")$", token)

        if not found:
            return None

        direction_min = int(found[1])
        direction_max = int(found[2])

        direction_min = convert_direction(direction_min)
        direction_max = convert_direction(direction_max)

        return VariableDirection(
            direction_min=direction_min, direction_max=direction_max
        )


@dataclass_json
@dataclass
class Visibility:
    """
    Visibility conditions
    Distance is expressed in meters
    """

    distance: Number
    cavok: Optional[bool]

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the visibility conditions from a token

        :param token: token
        :return: visibility conditions instance or None if not successful
        """
        found = re.search(
            "^(CAVOK|([0-9]{4})|([PM])?([0-9]{0,2})?(([1357])/(2|4|8|16))?SM|////)$",
            token,
        )

        if not found:
            return None

        cavok = False

        # get visibility and cavok flag
        if found[1] == "CAVOK":
            cavok = True
            distance = 9999
        elif found[1] == "////":
            distance = None
        else:
            if found[2]:
                distance = int(found[2])
            else:
                distance = 0 if not found[4] else int(found[4])

                if found[7]:
                    distance += int(found[6]) / int(found[7])

                distance = convert_distance(distance, "SM")

            distance = min(distance, 9999)

        return cls(distance=distance, cavok=cavok)


@dataclass_json
@dataclass
class Weather:
    """
    Weather conditions
    """

    intensity: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    descriptor: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    phenomena: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the weather conditions from a token

        :param token: token
        :return: weather conditions instance or None if not successful
        """
        expression = (
                "^([-+]|VC)?("
                + WEATHER_DESCRIPTOR
                + ")?("
                + WEATHER_PHENOMENA
                + ")?("
                + WEATHER_PHENOMENA
                + ")?("
                + WEATHER_PHENOMENA
                + ")?(NSW)?$"
        )
        found = re.search(expression, token)

        if not found:
            return None

        intensity = None
        descriptor = None
        phenomena = None

        # intensity
        if found[1]:
            intensity = found[1]

        # descriptor
        if found[2]:
            descriptor = found[2]

        # for all codes after intensity ...
        if found[3]:
            phenomena = ""
            for code in found.groups()[2:]:
                if code:
                    phenomena += code

        return cls(
            intensity=intensity,
            descriptor=descriptor,
            phenomena=phenomena,
        )


@dataclass_json
@dataclass
class Clouds:
    """
    Cloud conditions
    Height expressed in meters
    """

    amount: Optional[Union[str, Number]] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    height: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    vertical_visibility: Optional[Number] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )
    cloud: Optional[str] = field(
        default=None, metadata=config(exclude=lambda x: x is None)
    )

    @classmethod
    def from_text(cls, token: str):
        """
        Attempts to parse the cloud conditions from a token

        :param token: token
        :return: cloud conditions instance or None if not successful
        """
        found = re.search(
            "^((NSC|NCD|CLR|SKC|NOBS)|((VV|FEW|SCT|BKN|OVC|///)([0-9]{3}|///)(CB|TCU|///)?))$",
            token,
        )

        if not found:
            return None

        amount = None
        height = None
        vertical_visibility = None
        cloud = None

        # get cloud height and amount
        if found[2] and found[2] in CloudCover.set():
            amount = found[2]
        elif found[5]:
            if found[5] != "///":
                height = convert_distance(int(found[5]) * 100, "FT")
            else:
                # unknown height
                height = None

            if found[4] in CloudCover.set():
                amount = found[4]

                # VV is an special case
                if amount == "VV":
                    vertical_visibility = height
                    height = None 

        # get cloud type
        if found[6] in CloudType.set():
            cloud = found[6]

        return cls(amount=amount, height=height, vertical_visibility=vertical_visibility, cloud=cloud)


def clouds_height(clouds: Union[List[Clouds], None]) -> Union[Number, None]:
    """
    Computes the ceiling based on cloud layers
    A cloud ceiling is the height of the first cloud layer that constitutes at least a broken (BKN) layer

    :param clouds: layers of clouds
    :return: the ceiling in meters
    """
    ceiling = 3048

    if clouds is None:
        return ceiling

    for c in clouds:
        if c.amount in ["BKN", "OVC"]:
            if c.height is not None:
                ceiling = min(ceiling, c.height)

    return ceiling


def clouds_cloud(clouds: Union[List[Clouds], None]) -> Union[str, None]:
    """
    Computes the most dangerous cloud (TCU or CB) based on cloud layers

    :param clouds: layers of clouds
    :return: the most dangerous cloud (if any)
    """
    if clouds is None:
        return None

    members = CloudType.list()
    idx = None
    for c in clouds:
        if c.cloud is not None:
            idx = max(members.index(c.cloud), 0 if idx is None else idx)

    if idx is not None:
        return members[idx]
    else:
        return None


def clouds_amount(clouds: Union[List[Clouds], None]) -> Union[Number, None]:
    """
    Computes the largest layer cover in oktas

    :param clouds: layers of clouds
    :return: the most restrictive (maximum) amount of clouds in oktas
    """
    if clouds is None:
        return None

    members = CloudCover.list()
    idx = None
    
    for c in clouds:
        if c.amount is not None and c.amount != "VV":
            idx = max(members.index(c.amount), 0 if idx is None else idx)

    if idx is not None:
        return (idx + 1) * 2
    else:
        return None


def clouds_vertical_visibility(clouds: Union[List[Clouds], None]) -> Union[Number, None]:
    """
    Computes the vertical_visibility from layers of cloudss

    :param clouds: layers of clouds
    :return: the most restrictive (minimum) vertical visibilitys
    """
    if clouds is None:
        return None

    vertical_visibility = None

    for c in clouds:
        if c.amount == "VV" and c.vertical_visibility is not None:
            vertical_visibility = min(c.vertical_visibility, float(
                "inf") if vertical_visibility is None else vertical_visibility)

    return vertical_visibility
    

def simplify_clouds(clouds: List[Clouds]) -> Clouds:
    """
    Simplifies list of clouds

    :param clouds: layers of clouds
    :return: simplified clouds with the most restrictive information
    """
    return Clouds(
        height=clouds_height(clouds),
        vertical_visibility=clouds_vertical_visibility(clouds),
        cloud=clouds_cloud(clouds),
        amount=clouds_amount(clouds),
    )



#!/usr/bin/env python
# coding: utf-8
"""
metafora - enumerators
"""
from enum import Enum


class CustomEnum(str, Enum):
    """
    Custom enum with additional functionalities
    """

    @classmethod
    def set(cls):
        """
        Returns the members of this enumerator

        :return: members
        """
        return set(k for k in cls.__members__.keys())

    @classmethod
    def list(cls):
        """
        Returns the members of this list

        :return: members
        """
        return list(k for k in cls.__members__.keys())


class CloudCover(CustomEnum):
    """
    Enumerator of cloud covers
    """
    FEW = "a few"
    SCT = "scattered"
    BKN = "broken sky"
    OVC = "overcast sky"
    VV = "vertical visibility"


class CloudType(CustomEnum):
    """
    Enumerator of cloud types
    """
    TCU = "towering cumulus"
    CB = "cumulonimbus"


class WeatherDescriptor(CustomEnum):
    """
    Enumerator of weather descriptors
    """
    MI = "shallow"
    PR = "partial"
    BC = "patches of"
    DR = "low drifting"
    BL = "blowing"
    SH = "showers of"
    TS = "thunderstorms"
    FZ = "freezing"


class WeatherPrecipitation(CustomEnum):
    """
    Enumerator of weather precipitation
    """
    DZ = "drizzle"
    RA = "rain"
    SN = "snow"
    SG = "snow grains"
    IC = "ice crystals"
    PL = "ice pellets"
    GR = "hail"
    GS = "small hail"
    UP = "unknown"


class WeatherObscuration(CustomEnum):
    """
    Enumerator of weather obscuration
    """
    BR = "mist"
    FG = "fog"
    FU = "smoke"
    DU = "widespread dust"
    SA = "sand"
    HZ = "haze"


class OtherWeather(CustomEnum):
    """
    Enumerator of other weather phenomena
    """
    PY = "spray"
    VA = "volcanic ash"
    PO = "well-developed dust/sand whirls"
    SQ = "squalls"
    FC = "funnel cloud, tornado, or waterspout"
    SS = "sandstorm"
    DS = "duststorm"


class RunwayVisualRangeTrend(CustomEnum):
    """
    Enumerator of event tendencies
    """
    D = "decreasing"
    U = "increasing"
    N = "no tendency"


class ChangeEu(CustomEnum):
    """
    Enumerator of change indicators
    """
    BECMG = "expected to arise soon"
    TEMPO = "expected to arise temporarily"
    INTER = "expected to arise intermittent"


class ChangeUS(CustomEnum):
    """
    Enumerator of time indicators
    """
    FM = "from"
    AT = "at"
    TL = "until"



#!/usr/bin/env python
# coding: utf-8
"""
metafora utilities
"""
from typing import Union

Number = Union[float, int]

COMPASS = [
            "N",
            "NNE",
            "NE",
            "ENE",
            "E",
            "ESE",
            "SE",
            "SSE",
            "S",
            "SSW",
            "SW",
            "WSW",
            "W",
            "WNW",
            "NW",
            "NNW",
        ]

        
def convert_speed(speed: Number, unit: str) -> float:
    """
    Convets speed to meters per second

    :param speed: speed in the original units
    :param unit: units in KT, KPH or MPS
    :return: speed in meters per second
    """
    if unit == "KT":
        return round(0.514444 * speed, 2)
    elif unit == "KPH":
        return round(0.277778 * speed, 2)
    elif unit == "MPS":
        return round(speed, 2)
    else:
        raise ValueError("Unknown speed unit {}. Speed unit must be KT, KPH or MPS".format(unit))


def convert_distance(distance: Number, unit: str) -> float:
    """
    Converts distance to meters

    :param distance: distance in the original units
    :param unit: units in FT, SM or M
    :return: distance in meters
    """
    if unit == "FT":
        return round(0.3048 * distance)
    elif unit == "SM":
        return round(1609.34 * distance)
    elif unit == "M":
        return round(distance)
    else:
        raise ValueError("Unknown distance unit {}. Distance unit must be FT, SM or M".format(unit))


def convert_pressure(pressure: Number, unit: str) -> float:
    """
    Converts pressure to hPa

    :param pressure: pressure in the original units
    :param unit: units in A or Q
    :return: pressure in hPa
    """
    if unit == "A":
        return round(pressure / 0.02953)
    elif unit == "Q":
        return round(pressure)
    else:
        raise ValueError("Unknown pressure unit {}. Pressure unit must be A or Q".format(unit))


def convert_direction(direction: Number) -> str:
    """
    Returns the compass of a wind direction in degrees

    :param direction: direction in degrees
    :return: compass
    """
    if 0 <= direction <= 360:
        return COMPASS[round(direction / 22.5) % 16]
    else:
        raise ValueError(
            "Wind direction must be in degrees. Thus, its value must be higher than 0 and lower than 360, "
            "while the provided value is {}".format(direction)
        )



#!/usr/bin/env python
from setuptools import setup

setup(name='metafora',
      version="1.1.5",
      description='METAR and TAF parser with dataclasses and regular expressions',
      author='Ramon Dalmau-Codina',
      author_email='ramon.dalmau.codina@gmail.com',
      url='https://github.com/ramondalmau/metafora.git',
      packages=["metafora"],
      python_requires='>=3.7',
      install_requires=[
          'dataclasses>=0.6',
          'marshmallow>=3.17.0',
          'dataclasses-json>=0.5.7']
      )
